<?php

namespace Drupal\theme_compiler_scss\Plugin\ThemeCompiler;

use Drupal\Core\Cache\CacheableResponse;
use Drupal\Core\Extension\ThemeHandlerInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Theme\ThemeInitializationInterface;
use Drupal\Core\Theme\ThemeManagerInterface;

use Drupal\theme_compiler\Plugin\ThemeCompilerPluginBase;
use Drupal\theme_compiler\Plugin\ThemeCompilerTargetContext;

use ScssPhp\ScssPhp\Compiler;

use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Defines the SCSS theme compiler plugin.
 *
 * @ThemeCompiler("scss_compiler")
 */
class ScssCompiler extends ThemeCompilerPluginBase implements ContainerFactoryPluginInterface {

  /**
   * The theme handler service.
   *
   * @var \Drupal\Core\Extension\ThemeHandlerInterface
   */
  protected $themeHandler;

  /**
   * The theme initialization service.
   *
   * @var \Drupal\Core\Theme\ThemeInitializationInterface
   */
  protected $themeInitialization;

  /**
   * The theme manager service.
   *
   * @var \Drupal\Core\Theme\ThemeManagerInterface
   */
  protected $themeManager;

  /**
   * {@inheritdoc}
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, ThemeHandlerInterface $theme_handler, ThemeInitializationInterface $theme_initialization, ThemeManagerInterface $theme_manager) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);

    $this->themeHandler = $theme_handler;
    $this->themeInitialization = $theme_initialization;
    $this->themeManager = $theme_manager;
  }

  /**
   * Passes alterable variables to enabled $theme_TYPE_alter() implementations.
   *
   * This method operates much like the theme manager's ::alter() method,
   * however this method will allow all enabled themes to alter the arguments.
   *
   * @param string $type
   *   A string describing the type of the alterable $data.
   * @param mixed $data
   *   The variable that will be passed to $theme_TYPE_alter() implementations
   *   to be altered.
   * @param mixed $context1
   *   An additional variable that is passed by reference (default: NULL).
   * @param mixed $context2
   *   An additional variable that is passed by reference (default: NULL).
   *
   * @see \Drupal\Core\Theme\ThemeManagerInterface::alter()
   *   For more information about the intended behavior of this method.
   */
  protected function alter($type, &$data, &$context1 = NULL, &$context2 = NULL) {
    // Allow themes to alter the compiler before execution.
    //
    // Normally only the current theme is given an opportunity to alter the
    // compilation process, so we have to iterate over all enabled themes
    // manually to give them an opportunity to modify compilation.
    //
    // @see \Drupal\Core\Theme\ThemeManagerInterface::alter()
    // @see \Drupal\Core\Theme\ThemeManagerInterface::alterForTheme()
    foreach ($this->themeHandler->listInfo() as $name => $extension) {
      // Fetch an "active theme" instance for this theme.
      //
      // This might be a bit of a misnomer in Drupal core; the theme isn't
      // necessarily "active" (i.e., visible on the current page).
      $active_theme = $this->themeInitialization->initTheme($name);

      // Allow this theme to alter the compilation process.
      $this->themeManager->alterForTheme($active_theme, $type, $data, $context1, $context2);
    }
  }

  /**
   * Generate the output for the compiler target using the provided options.
   *
   * @param \Drupal\theme_compiler\Plugin\ThemeCompilerTargetContext $context
   *   The compiler context containing common properties used by most (if not
   *   all) compilers.
   *
   * @return string
   *   The resulting output of the compiler.
   */
  protected function compile(ThemeCompilerTargetContext $context) {
    // Retrieve the theme being used and options from this context.
    $theme   = $context->getTheme();
    $options = $context->getOptions();

    // Create an instance of the SCSS compiler.
    $scss = new Compiler();

    // Add import paths as configured in the theme.
    $scss->setImportPaths($this->getImportPaths($theme, $options));
    // Set any variables collected from hook implementations.
    $scss->setVariables($this->getVariables($theme));

    // Provide the opportunity to alter the compiler before it's ran.
    $this->alter('theme_compiler_scss', $scss);

    // Ensure that a valid type is being used for 'import_paths'.
    if (!($scss instanceof Compiler)) {
      throw new \InvalidArgumentException('Bad type for SCSS compiler');
    }

    // Compile the contents of the source file(s) defined in the theme.
    $source = implode("\r\n", array_map(function ($path) {
      // Attempt to load the source file; return an empty string on failure.
      return file_get_contents($path) ?: '';
    }, $this->getSourcePaths($theme, $options)));

    return $scss->compile($source);
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('theme_handler'),
      $container->get('theme.initialization'),
      $container->get('theme.manager'),
    );
  }

  /**
   * Generate an absolute file system path for a theme-relative path segment.
   *
   * @param string $theme
   *   The machine name of the theme used to resolve relative paths.
   * @param string $segment
   *   The theme-relative path segment to resolve to an absolute path.
   *
   * @return string|bool
   *   An absolute file system path or FALSE if inexistent.
   */
  protected function getAbsoluteThemePath(string $theme, string $segment) {
    // Retrieve the requested theme's extension instance.
    $theme = $this->themeHandler->getTheme($theme);
    // Construct an absolute file system path using the theme-relative segment.
    return realpath($theme->getPathInfo()->getPathname() . '/' . $segment);
  }

  /**
   * Retrieve the final list of import paths after both hook invocations.
   *
   * The import paths are first loaded from the theme compiler configuration
   * then `hook_theme_compiler_scss_relimport_alter()` is called. The resulting
   * list of theme-relative import paths is then resolved to absolute file
   * system paths.
   *
   * Finally, `hook_theme_compiler_scss_absimport_alter()` is called to allow
   * extensions to include resources external to the configured theme.
   *
   * @param string $theme
   *   The machine name of the theme used to resolve relative paths.
   * @param array $options
   *   An array of compiler options.
   *
   * @return string[]
   *   An array of absolute file system paths used to import resources during
   *   SCSS compilation.
   */
  protected function getImportPaths(string $theme, array $options) {
    // Retrieve the import paths defined in the theme compiler config.
    $import_paths = $this->getThemeImportPaths($options);

    // Allow modules and themes to alter the list of relative import paths.
    $this->alter('theme_compiler_scss_relimport', $import_paths);

    // Process the resulting import paths.
    $import_paths = array_map(function ($path) use ($theme) {
      // Resolve this path to an absolute file system path.
      return $this->getAbsoluteThemePath($theme, $path);
    }, is_array($import_paths) ? $import_paths : []);

    // Allow modules and themes to alter the list of absolute import paths.
    $this->alter('theme_compiler_scss_absimport', $import_paths);

    return is_array($import_paths) ? $import_paths : [];
  }

  /**
   * Retrieve the final list of source paths after both hook invocations.
   *
   * The source paths are first loaded from the theme compiler configuration
   * then `hook_theme_compiler_scss_relsource_alter()` is called. The resulting
   * list of theme-relative source paths is then resolved to absolute file
   * system paths.
   *
   * Finally, `hook_theme_compiler_scss_abssource_alter()` is called to allow
   * extensions to include resources external to the configured theme.
   *
   * @param string $theme
   *   The machine name of the theme used to resolve relative paths.
   * @param array $options
   *   An array of compiler options.
   *
   * @return string[]
   *   An ordered set of absolute file system paths to SCSS source files.
   */
  protected function getSourcePaths(string $theme, array $options) {
    // Retrieve the source paths defined in the theme compiler config.
    $source = $this->getThemeSourcePaths($options);

    // Allow modules and themes to alter the list of relative source paths.
    $this->alter('theme_compiler_scss_relsource', $source);

    // Process the resulting source paths.
    $source = array_map(function ($path) use ($theme) {
      // Resolve this path to an absolute file system path.
      return $this->getAbsoluteThemePath($theme, $path);
    }, is_array($source) ? $source : []);

    // Allow modules and themes to alter the list of absolute source paths.
    $this->alter('theme_compiler_scss_abssource', $source);

    return is_array($source) ? $source : [];
  }

  /**
   * Attempts to load a set of import paths defined in the compiler config.
   *
   * If no import paths have been defined, an empty array is returned.
   *
   * @param array $options
   *   An array of compiler options.
   *
   * @throws \InvalidArgumentException
   *   When an invalid data type is provided for 'import_paths'.
   *
   * @return string[]
   *   An array of theme-relative import paths.
   */
  protected function getThemeImportPaths(array $options) {
    // Attempt to retrieve any import paths defined in theme compiler config.
    if (array_key_exists('import_paths', $options)) {
      $import_paths = $options['import_paths'];
    }
    else {
      $import_paths = [];
    }

    // Ensure that a valid type is being used for 'import_paths'.
    if (!is_array($import_paths)) {
      throw new \InvalidArgumentException('"import_paths" must be an array');
    }

    return $import_paths;
  }

  /**
   * Retrieve a list of source paths defined in the compiler config.
   *
   * @param array $options
   *   An array of compiler options.
   *
   * @throws \RuntimeException
   *   If no source has been defined in the compiler config.
   * @throws \InvalidArgumentException
   *   If 'source' is not a string or array.
   *
   * @return string[]
   *   An array of theme-relative source paths.
   */
  protected function getThemeSourcePaths(array $options) {
    // Make sure a source has been defined in the compiler options.
    if (array_key_exists('source', $options)) {
      $source = $options['source'];
    }
    else {
      throw new \RuntimeException('"source" is missing in compiler config');
    }

    // Ensure that a valid type is being used for 'source'.
    if (!is_string($source) && !is_array($source)) {
      throw new \InvalidArgumentException('"source" must be a path or a list of paths');
    }

    // Convert a single source into an array of sources.
    return is_array($source) ? array_filter($source, 'is_string') : [$source];
  }

  /**
   * Allow extensions to define SCSS variables before compilation.
   *
   * @param string $theme
   *   The machine name of the theme.
   *
   * @return array
   *   An associative array of variables to set before compilation begins.
   */
  protected function getVariables(string $theme) {
    $variables = [];

    // Allow modules and themes to alter the list of relative import paths.
    $this->alter('theme_compiler_scss_variables', $variables);

    // Return the resulting list of variables.
    $variables = is_array($variables) ? $variables : [];
    return array_filter($variables, 'is_string', ARRAY_FILTER_USE_KEY);
  }

  /**
   * {@inheritdoc}
   */
  public function serve(ThemeCompilerTargetContext $context) {
    // Run the compiler with the extracted options from this target context.
    $compiled = $this->compile($context);

    // Create a response using the compiled output and configure its
    // cacheability metadata appropriately.
    return new CacheableResponse($compiled, CacheableResponse::HTTP_OK, [
      'Content-Type' => 'text/css',
    ]);
  }

}
